package checkHadoop;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.net.URI;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Iterator;
import java.util.Map.Entry;
import java.util.Set;
import java.util.StringTokenizer;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.IntWritable;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.Text;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Reducer;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;


public class ActionRules{
	public static class JobMapper extends Mapper<LongWritable, Text, Text, Text>{

		int lineCount = 0;

		boolean falseParameters;
		
		double minSupport,minConfidence;
		
		String decisionAttribute,decisionFrom,decisionTo;
		
		public static ArrayList<String> attributeNames,stableAttributes,stableAttributeValues;
		public static ArrayList<ArrayList<String>> actionRules;
		
		static Map<ArrayList<String>, Integer> data;
		static Map<String, HashSet<String>> distinctAttributeValues,decisionValues;
		static Map<HashSet<String>, HashSet<String>> attributeValues,reducedAttributeValues;
		static Map<ArrayList<String>, HashSet<String>> markedValues,possibleRules;
		static Map<ArrayList<String>,String> certainRules;
		
		public JobMapper() {
			super();
			falseParameters = false;
			
			data = new HashMap<ArrayList<String>, Integer>();
			actionRules = new ArrayList<ArrayList<String>>();
			attributeNames = new ArrayList<String>();
			stableAttributes = new ArrayList<String>();
			stableAttributeValues = new ArrayList<String>();
		
			attributeValues = new HashMap<HashSet<String>, HashSet<String>>();
			reducedAttributeValues = new HashMap<HashSet<String>, HashSet<String>>();
			distinctAttributeValues = new HashMap<String, HashSet<String>>();
			decisionValues = new HashMap<String, HashSet<String>>();
			certainRules = new HashMap<ArrayList<String>, String>();
			markedValues = new HashMap<ArrayList<String>, HashSet<String>>();
			possibleRules = new HashMap<ArrayList<String>, HashSet<String>>();
		}
		
		/*
		 * 
		 * setup() function is to configure all mappers with the common data.
		 * In our case, it is attributes and user parameters.
		 * Since these values has to be given completely to all maps, we are using 
		 * 				them as global values for all maps. 
		 * 
		 */
		
		@Override
		protected void setup(
				Mapper<LongWritable, Text, Text, Text>.Context context)
				throws IOException, InterruptedException {
			/*
			 * Stored Cache files are accessed here!!!
			 * The function used here is deprecated. I'm not able to find an alternative
			 * */
			@SuppressWarnings("unused")
			Path[] files = context.getLocalCacheFiles();
			
			for (Path path : files) {
				
				//Reading Attribute file from the Distributed Cache
				if (path.getName().equals("attributes.txt")) {
					BufferedReader reader = new BufferedReader(new FileReader(path.toString()));
					String fileLine = reader.readLine();
					
					while(fileLine!=null){
						String[] tokens = fileLine.split("\\s+");
                                                
                                                for (int i=0; i<tokens.length; i++)
                                                {
                                                    context.write(new Text("Attributes"), new Text("Attributes - "+tokens[i]));
                                                    attributeNames.add(tokens[i]);
                                                    
                                                }
                                                
                                                
                                                
                                                //attributeNames.add(tokens[0]);
                                                //context.write(new Text("Attributes"), new Text("Attributes - "+tokens[1]));
						fileLine = reader.readLine();
					}
				}
				
				//Reading User Parameters file from the Distributed Cache
				else if(path.getName().equals("parameters.txt")){
					BufferedReader reader = new BufferedReader(new FileReader(path.toString()));
					String fileLine = reader.readLine();
					int counter = 0;
					
					while(fileLine!=null){
						counter++;
						
						StringTokenizer tokens = new StringTokenizer(fileLine);
						
						switch(counter){
							case 1://Line 1 in parameters.txt to get the stable attribute names
								while(tokens.hasMoreTokens()){
									String attributeName = tokens.nextToken();
                                                                        context.write(new Text("Stable Attribute"), new Text(attributeName));
									if(checkInStringList(attributeNames, attributeName)){
										stableAttributes.add(attributeName);
									} else falseParameters = true;
								}
								break;
								
							case 2://Line 2 in parameters.txt to get the decision attribute name
								if(tokens.countTokens() == 1){
									String attributeName = tokens.nextToken();
                                                                        context.write(new Text("Decision Attribute"), new Text(attributeName));
									if(checkInStringList(attributeNames, attributeName)){
										decisionAttribute = attributeName;
                                                                                //context.write(new Text("Decision attribute if:"), new Text("Decision att:"+decisionAttribute));
									} else falseParameters = true;
								} else falseParameters = true;
								break;
								
							case 3://Line 3 in parameters.txt to get decisionFrom and decisionTo attribute values
								if(tokens.countTokens() == 2){
									decisionFrom = decisionAttribute +"_"+ tokens.nextToken();
									decisionTo = decisionAttribute +"_"+ tokens.nextToken();
                                                                        
                                                                        context.write(new Text("From -> To"), new Text(decisionFrom + " -> "+decisionTo));
								} else falseParameters = true;
								break;
								
							case 4://Line 4 in parameters.txt to get Support and Confidence
								if(tokens.countTokens() == 2){
									minSupport = Double.parseDouble(tokens.nextToken());
									minConfidence = Double.parseDouble(tokens.nextToken());
                                                                        
                                                                        context.write(new Text("minSupport -> minConfidence"), new Text(minSupport + " -> "+minConfidence));
                                                                        
								} else falseParameters = true;
								break;
								
						}
						
						fileLine = reader.readLine();
					}
				}
			}
			
			super.setup(context);
		}
		
		//Check if the value is present in the given list
		private boolean checkInStringList(ArrayList<String> list, String value){
			if(list.contains(value)){
				return true;
			}
			return false;
		}


		/*
		 * 
		 * map() function reads only the data from the Input file
		 * We used in this way because map() functions runs algorithm for each line of the data
		 * Since LERS and ARoGS can be performed in a group of data, we are using this function 
		 * 				to read data and put them in separate Data Structures
		 * 
		 */
		@Override
		protected void map(
				LongWritable key,
				Text inputValue,
				Mapper<LongWritable, Text, Text, Text>.Context context)
				throws IOException, InterruptedException {
                    
                                            
			
			if(falseParameters){
				Text checkText = new Text("Error:");
				HashSet<String> errorMessage = new HashSet<String>();
				errorMessage.add("Invalid Parameters");
				
			}else{
                            
                                //context.write(new Text("Splitting"), new Text(inputValue));
				splitData(inputValue,lineCount, context);
				
                                
				//Getting stable attribute values
				for(int i = 0;i<stableAttributes.size();i++){
					HashSet<String> distinctStableValues = distinctAttributeValues.get(stableAttributes.get(i));
					
					for (String string : distinctStableValues) {
                                            if(!stableAttributeValues.contains(string))
                                                    stableAttributeValues.add(string);
                                            else continue;
					}
					
				}
				
				setDecisionAttributeValues();
                                
                                
			}
			
			
			lineCount++;
                    
			super.map(key, inputValue, context);
		}

		//Splitting the input data
		private void splitData(Text inputValue, int lineCount, Mapper<LongWritable, Text, Text, Text>.Context context) throws IOException, InterruptedException {
			int lineNo = lineCount;
				
			String inputData = inputValue.toString();

			//This line splits only tab separated or comma separated data
			ArrayList<String> lineData = new ArrayList<String>(Arrays.asList(inputData.split("\t|,|\\s+")));
			
                        
			if(!checkEmptyValueInStringArray(lineData) ){//Omitting the data line, if it contains empty value
				String key;
				
				lineNo++;
				
				ArrayList<String> tempList = new ArrayList<String>();
				HashSet<String> set;
					
				for (int j=0;j<lineData.size();j++) {
					//Creating proper attribute value like Aa1,Bb1 
					//since it would be easy to process them later in the algorithm
					String currentAttributeValue = lineData.get(j);
					String attributeName = attributeNames.get(j);
                                        key = attributeName +"_"+ currentAttributeValue;//attributeName.concat(currentAttributeValue);
                                        
                                        //context.write(new Text("Split Data - "), new Text(key));
						
					tempList.add(key);
								
					if (distinctAttributeValues.containsKey(attributeName)) {
						set = distinctAttributeValues.get(attributeName);
						
					}else{
						set = new HashSet<String>();
					}
							
					set.add(key);
					//Setting attribute values to the corresponding attribute
					distinctAttributeValues.put(attributeName, set);
				}
							
				if(!data.containsKey(tempList)){
					data.put(tempList,1);
				
					for(String listKey : tempList){
						HashSet<String> mapKey = new HashSet<String>();
						mapKey.add(listKey);
						setMap(attributeValues,mapKey,lineNo);
					}
				}
				else
					data.put(tempList, data.get(tempList) + 1);
			} 
					
		}
		
		
		//Checks if string array contains empty value
		private static boolean checkEmptyValueInStringArray(ArrayList<String> lineData) {
                        
			return lineData.contains("");
                        
		}
				
		//Sets Attribute values - similar lineNumber
		private static void setMap(Map<HashSet<String>, HashSet<String>> values,
				HashSet<String> key, int lineNo) {
			HashSet<String> tempSet = new HashSet<String>();
			
			if (values.containsKey(key)) {
				tempSet.addAll(values.get(key));						
			}
			
			tempSet.add("x"+lineNo);
			values.put(key, tempSet);
		}
		
		//Sets Decision attribute values
		private void setDecisionAttributeValues() {
			HashSet<String> distinctDecisionValues = distinctAttributeValues.get(decisionAttribute);
			for(String value : distinctDecisionValues){
				HashSet<String> newHash = new HashSet<String>();
				HashSet<String> finalHash = new HashSet<String>();
				newHash.add(value);
				
				if(decisionValues.containsKey(value)){
					finalHash.addAll(decisionValues.get(value));
				}
				if(attributeValues.containsKey(newHash))
					finalHash.addAll(attributeValues.get(newHash));
				
				decisionValues.put(value, finalHash);
				attributeValues.remove(newHash);
			}
		}
		
		/*
		 * This function implements both LERS and ARoGS algorithms
		 * */
		@Override
		protected void cleanup(
				final Mapper<LongWritable, Text, Text, Text>.Context context)
				throws IOException, InterruptedException {
                    
                            performLERS(context);
			
                            generateActionRules(context); 
                            /*
                            for(Map.Entry<ArrayList<String>,Integer> entry : data.entrySet()){
                                context.write(new Text("Check Data"), new Textdata(entry.getKey().toString()));
                            }*/
                        
			super.cleanup(context);
		}
		
		//Finding LERS rules
		private void performLERS(Mapper<LongWritable, Text, Text, Text>.Context context) throws IOException, InterruptedException {
                        
                        context.write(new Text("LERS - "), new Text("called"));
                    
			int loopCount = 0;
			
			while(!attributeValues.isEmpty()){

				for (Map.Entry<HashSet<String>, HashSet<String>> set : attributeValues.entrySet()) {
					ArrayList<String> setKey = new ArrayList<String>();
					setKey.addAll(set.getKey());
					
					HashSet<String> setValue = set.getValue();
					
					if (set.getValue().isEmpty()) {
						continue;
					}else{
						for(Map.Entry<String, HashSet<String>> decisionSet : decisionValues.entrySet()){
                                                    if(decisionSet.getValue().containsAll(set.getValue()) && calculateSupportLERS(setKey,decisionSet.getKey())!=0){
							if(decisionSet.getValue().containsAll(set.getValue())){
								certainRules.put(setKey, decisionSet.getKey());
								markedValues.put(setKey, set.getValue());
								break;
							}
                                                    }
                                                
						}
						
					}
							
					if(!markedValues.containsKey(setKey)){
						HashSet<String> possibleRulesSet = new HashSet<String>();
						for(Map.Entry<String, HashSet<String>> decisionSet : decisionValues.entrySet()){
//							if(calculateSupportLERS(setKey,decisionSet.getKey())!=0)
								possibleRulesSet.add(decisionSet.getKey());
						}
								
						if(possibleRulesSet.size()>0)
							possibleRules.put(setKey, possibleRulesSet);
					}
				}
				
				removeMarkedValues();
				
				combinePossibleRules();
			}
			
		}
		
		
		//Finding Support and Confidence for LERS rules
		static int calculateSupportLERS(ArrayList<String> key, String value) {
			ArrayList<String> tempList = new ArrayList<String>();
			
			for(String val : key){
				if(!val.equals(""))
					tempList.add(val);
			}
					
			if(!value.equals(""))
				tempList.add(value);
		
			return findLERSSupport(tempList);
					
		}
				
		private static int findLERSSupport(ArrayList<String> tempList) {
			int count = 0;
				
			for(Map.Entry<ArrayList<String>, Integer> entry : data.entrySet()){
				if(entry.getKey().containsAll(tempList)){
					count += entry.getValue();
				}
			}
				
			return count;
		}
		
		//Finds Confidence
		static String calculateConfidenceLERS(ArrayList<String> key,
				String value) {
			int num = calculateSupportLERS(key, value);
			int den = calculateSupportLERS(key, "");
			int confidence = 0;
			
			if(den!=0){
				confidence = (num * 100)/den;
			}
//				else{
//				commonMethods.printMessage("------------->Check this");
//			}
			
			return String.valueOf(confidence);
			
		}
		
		//Removing marked sets
		private static void removeMarkedValues() {
			for(Map.Entry<ArrayList<String>, HashSet<String>> markedSet : markedValues.entrySet()){
				attributeValues.remove(new HashSet<String>(markedSet.getKey()));
			}
			
		}
				
		//Combining Possible Rules
		private static void combinePossibleRules() {
			Set<ArrayList<String>> keySet = possibleRules.keySet();
			ArrayList<ArrayList<String>> keyList = new ArrayList<ArrayList<String>>();
			keyList.addAll(keySet);
			
			for(int i = 0;i<possibleRules.size();i++){
				for(int j = (i+1);j<possibleRules.size();j++){
					HashSet<String> combinedKeys = new HashSet<String>(keyList.get(i));
					combinedKeys.addAll(new HashSet<String>(keyList.get(j)));
					
					if(!checkSameGroup(combinedKeys)){
						combineAttributeValues(combinedKeys);
					}
				}
			}
								
			removeRedundantValues();
			clearAttributeValues();
			possibleRules.clear();
					
		}
		
		//Checks if the values belong to the same attribute
		public static boolean checkSameGroup(HashSet<String> combinedKeys) {		
			ArrayList<String> combinedKeyAttributes = new ArrayList<String>();
			
			for(Map.Entry<String, HashSet<String>> singleAttribute : distinctAttributeValues.entrySet()){
				for(String key : combinedKeys){
					if(singleAttribute.getValue().contains(key)){
						if(!combinedKeyAttributes.contains(singleAttribute.getKey()))
								combinedKeyAttributes.add(singleAttribute.getKey());
						else return true;
					}
				}
			
			}
			
			return false;
		}
				
		//Performing Intersection of Sets
		private static void combineAttributeValues(HashSet<String> combinedKeys) {
			HashSet<String> combinedValues = new HashSet<String>();
					
			for(Map.Entry<HashSet<String>, HashSet<String>> attributeValue : attributeValues.entrySet()){
				if(combinedKeys.containsAll(attributeValue.getKey())){
					if(combinedValues.isEmpty()){
						combinedValues.addAll(attributeValue.getValue());
					}else{
						combinedValues.retainAll(attributeValue.getValue());
					}
				}
			}
			if(combinedValues.size()!=0)
				reducedAttributeValues.put(combinedKeys, combinedValues);
		}
				
		//Eliminating already marked values
		private static void removeRedundantValues() {
			HashSet<String> mark = new HashSet<String>();
			
			for(Map.Entry<HashSet<String>, HashSet<String>> reducedAttributeValue : reducedAttributeValues.entrySet()){
				for(Map.Entry<HashSet<String>, HashSet<String>> attributeValue : attributeValues.entrySet()){
					
					if(attributeValue.getValue().containsAll(reducedAttributeValue.getValue()) || reducedAttributeValue.getValue().isEmpty()){
						mark.addAll(reducedAttributeValue.getKey());
					}
				}
			}
				
			reducedAttributeValues.remove(mark);
			
		}
		
		private static void clearAttributeValues() {
			 attributeValues.clear();
			 for(Map.Entry<HashSet<String>, HashSet<String>> reducedAttributeValue : reducedAttributeValues.entrySet()){
				 attributeValues.put(reducedAttributeValue.getKey(), reducedAttributeValue.getValue());
			 }
			 reducedAttributeValues.clear();
		}

		static Map<String,ArrayList<String>> atomicActionMap = new HashMap<String, ArrayList<String>>();
		
		//Generates Action Rules
		public ArrayList<String> generateActionRules(Mapper<LongWritable, Text, Text, Text>.Context context) throws IOException, InterruptedException {
			
                        //context.write(new Text("Generate Action Rules - "), new Text("called"));
                        ArrayList<String> actionFrom, actionTo, rules;
			ArrayList<String> actions = null;
		
			String rule = "";
			String action = "A";
			int count = 0;
			
			rules = new ArrayList<String>();
                        //for printing certain rules

                        for(Map.Entry<ArrayList<String>, String> certainRules1 : certainRules.entrySet()){
				ArrayList<String> list = new ArrayList<String>();
				if(certainRules1.getValue().equals(decisionFrom)){
					for(Map.Entry<ArrayList<String>, String> certainRules2 : certainRules.entrySet()){
						if((certainRules1.getKey().equals(certainRules2.getKey())) || (!certainRules2.getValue().equals(decisionTo))
								|| (certainRules1.getKey().size()>=certainRules2.getKey().size() )) continue;
						else{
							String primeAttribute = "";
							ArrayList<String> checkCertainValues1 = certainRules1.getKey();
							ArrayList<String> checkCertainValues2 = certainRules2.getKey();
							//int support1 = calculateSupportLERS(checkCertainValues1,certainRules1.getValue());
							//int support2 = calculateSupportLERS(checkCertainValues2,certainRules2.getValue());
							//int minSup = support1<support2?support1:support2;

							//float conf1 = findConfidence(support1, checkCertainValues1);
							//float conf2 = findConfidence(support2, checkCertainValues2);
							//float confidence = calculateConfidence(conf1, conf2);

	//						if(minSup>=supportThreshold || confidence>=confiThreshold){

								for(String value1 : checkCertainValues1){
                                                                        String[] val1 = value1.split("_");
									if(stableAttributes.contains(val1[0])) {
                                                                            continue;
                                                                        }
									else{
										primeAttribute = checkAttribute(value1,context);
                                                                                if(!checkCertainValues2.contains("?")){
										for(String value2 : checkCertainValues2){
                                                                                        
                                                                                        //trial code
                                                                                        String[] val2 = value2.split("_");
											
                                                                                        if(stableAttributes.contains(val2[0]) || !(checkAttribute(value2,context).equals(primeAttribute)))
                                                                                        {
                                                                                            
                                                                                                rule=formRule(rule,value2,value2,context);
                                                                                            
                                                                                        }
											else if(checkAttribute(value2,context).equals(primeAttribute))
												rule=formRule(rule,value1,value2,context);
										}
                                                                             }
									}
								}
//							}

							if(rule.indexOf(primeAttribute)!=-1 && !primeAttribute.isEmpty()){
								count++;
								if(rule.contains("^")){

									String[] atomicActionArray = rule.split("\\^");
									int len = atomicActionArray.length;
									ArrayList<String> mapValue ;

									for(int i=0; i<len; i++){
										if(atomicActionMap.containsKey(atomicActionArray[i])) {
											mapValue = atomicActionMap.get(atomicActionArray[i]);
											mapValue.add(action+count);
										} 
										else {
											mapValue = new ArrayList<String>();
											mapValue.add(action+count);
										}
										atomicActionMap.put(atomicActionArray[i],mapValue);
									}

								}
								context.write(new Text(rule + " ==> " + "(" + decisionFrom + "->" + decisionTo +")"),null);
								
							}
							rule = "";
						}
						//loop2Count++;
					}
				}else continue;
			}
            
                        try {
							writeIntoJSON(context);
							
						} catch (JSONException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}
                        

			
			return rules;
		}
		private static void writeIntoJSON(Mapper<LongWritable, Text, Text, Text>.Context context )throws IOException, InterruptedException, JSONException  {

			Path pt=new Path("/user/kgosalia/KDD_Part3/ontology.txt");
            FileSystem fs = FileSystem.get(new Configuration());
            BufferedWriter br=new BufferedWriter(new OutputStreamWriter(fs.create(pt,true)));
                                       // TO append data to a file, use fs.append(Path f)
            String line;
            
			Map<String,Integer> countMap= new HashMap<String,Integer>();
			JSONArray jtarget = new JSONArray();
			JSONArray ja = new JSONArray();

			Set<String> set = new HashSet<String>();
			int size = atomicActionMap.size();
			context.write(new Text(""+size), null);
			int count=0;
			int counter=0;

			Iterator<Entry<String, ArrayList<String>>> iterator1 = atomicActionMap.entrySet().iterator();
			
			while(iterator1.hasNext()) {
				Entry<String, ArrayList<String>> entry = iterator1.next();
				
				
				
				for(String str:entry.getValue()){
					context.write(new Text("ENtry"+entry.getKey()), new Text(str));
					set.add(str);	
				}
			}

			JSONObject jobj = new JSONObject();
			jobj.put("name", "A");
			jobj.put("size", "60");
			ja.put(jobj);

			for(String str:set){
				count++;
				JSONObject jo = new JSONObject();
				jo.put("name", str);
				jo.put("size", "30");
				ja.put(jo);

				counter++;
				JSONObject jtar = new JSONObject();
				jtar.put("source", 0);
				jtar.put("target", count);
				jtarget.put(jtar);

			}
			Iterator<Entry<String, ArrayList<String>>> iterator2 = atomicActionMap.entrySet().iterator();
			while(iterator2.hasNext()) {
				Entry<String, ArrayList<String>> entry = iterator2.next();

				JSONObject jo = new JSONObject();
				jo.put("name", entry.getKey());
				jo.put("size", "10");
				ja.put(jo);
				countMap.put(entry.getKey(), counter + 1);
				counter++;
			}


			for(String str:set){

				List<String> keys =getValueFromKeys(atomicActionMap, str);
				String subStr=str.substring(1);
				for(String key : keys){
					JSONObject jtar = new JSONObject();
					jtar.put("source", Integer.valueOf(subStr));
					jtar.put("target", countMap.get(key));
					jtarget.put(jtar);
				}
			}

			JSONObject mainObj = new JSONObject();
			mainObj.put("nodes", ja);
			mainObj.put("links", jtarget);

			br.write(mainObj.toString(2) );
			//System.out.printf( "%s", mainObj.toString(2) );
			br.close();
			   
		}
		private static List<String> getValueFromKeys(Map<String, ArrayList<String>> atomicActionMap, String str ) {

			List<String> values=  new ArrayList<String>();
			Iterator<Entry<String, ArrayList<String>>> itr = atomicActionMap.entrySet().iterator();
			while(itr.hasNext()){
				Entry<String, ArrayList<String>> entry = itr.next();
				if(entry.getValue().contains(str)) values.add(entry.getKey());
			}
			return values;

		}       
                private static String checkAttribute(String value,Mapper<LongWritable, Text, Text, Text>.Context context)
				throws IOException, InterruptedException
                {
                    
                   
                    for(Map.Entry<String,HashSet<String>> entryValue : distinctAttributeValues.entrySet())
                    {
                   
                        if(entryValue.getValue().contains(value))
                        {
                            return entryValue.getKey();
                        }
                    }
                    return null;
                }
                
                public void  write(){
                    
                    try{
                            Path pt=new Path("/user/kgosalia/KDD_Part3/ontology.txt");
                            FileSystem fs = FileSystem.get(new Configuration());
                            BufferedWriter br=new BufferedWriter(new OutputStreamWriter(fs.create(pt,true)));
                                                       // TO append data to a file, use fs.append(Path f)
                            String line;
                            line="Test Test1 Test2";
                            System.out.println(line);
                            br.write(line);
                            br.close();
                    }catch(Exception e){
                            System.out.println("File not found");
                    }
            }
		private static boolean checkStableAttributes(ArrayList<String> key,
				ArrayList<String> key2) throws IOException, InterruptedException {
			List<String> stableAttributesList1 = new ArrayList<String>();
			List<String> stableAttributesList2 = new ArrayList<String>();
			
			for(String value : key){
				if(stableAttributeValues.contains(value))
					stableAttributesList1.add(value);
			}
			
			for(String value : key2){
				if(stableAttributeValues.contains(value))
					stableAttributesList2.add(value);
			}
			
			if(stableAttributesList2.containsAll(stableAttributesList1))
				return true;
			else return false;
		}
		
		private boolean checkRulesSubSet(
				ArrayList<String> certainRules1,
				ArrayList<String> certainRules2) {
			ArrayList<String> primeAttributes1 = new ArrayList<String>();
			ArrayList<String> primeAttributes2 = new ArrayList<String>();
			
			for (String string : certainRules1) {
				String attributeName = getAttributeName(string);
				
				if(!isStable(attributeName))
					primeAttributes1.add(attributeName);
			}
			
			for (String string : certainRules2) {
				String attributeName = getAttributeName(string);
				
				if(!isStable(attributeName))
					primeAttributes2.add(attributeName);
			}
			
			if(primeAttributes2.containsAll(primeAttributes1))
				return true;
			else return false;
		}
		
		public static String getAttributeName(String value1) {
			for(Map.Entry<String, HashSet<String>> entryValue : distinctAttributeValues.entrySet()){
				if(entryValue.getValue().contains(value1)){
					return entryValue.getKey();
				}
			}
			return null;
		}
		
		private static String formRule(String rule,String value1, String value2, Mapper<LongWritable, Text, Text, Text>.Context context) throws IOException, InterruptedException {
                    
                    
                    //context.write(new Text("Rule - "), new Text(rule));
                    //context.write(new Text("Value 1 - "), new Text(value1));
                    //context.write(new Text("Value 2 - "), new Text(value2));
			
//                        if(!rule.isEmpty())
//				rule += "^";
//			
//			rule += "(" + getAttributeName(value2) + "," + getAction(value1,value2) + ")";
//			return rule;
                    
                boolean missingValue = value1.contains("?") || value2.contains("?");
                 if(!missingValue){
                    if(!rule.isEmpty())
			rule += "^";

		if(value1.equals(value2))
		{
			String[] val = value1.split("_");
                        
                       // context.write(new Text("val[0]->"+val[0]), new Text("val[1]->"+val[1]));
			if(stableAttributes.contains(val[0])) 
                        {
                            rule += "(" + val[0] + "=" + val[1] + ")";
                        }
			else 
                        {
                            rule += "(" + val[0] + ",->" + val[1] + ")";
                        }
		}
                else{
                        if(!value2.isEmpty() && !value1.isEmpty())
                        {
                            String[] val1 = value1.split("_");
                            String[] val2 = value2.split("_");  
                            
                            if(!(stableAttributes.contains(val1[0]))) {
                                 rule += "(" + val1[0] +","+val1[1] +"->" + val2[1] + ")";
                            }
                           
                        }
                        //rule += "(" + val1[0] +","+val1[1] +"->" + val2[1] + ")";
		}
                    }
		return rule;
	}
		
		private static String getAction(String left,String right){
			return left + "->" + right;
		}
		
		private void printActionRule(final ArrayList<String> actionFrom,final ArrayList<String> actionTo,
				final ArrayList<String> actions,final String rule, final Mapper<LongWritable, Text, Text, Text>.Context context) 
						throws IOException, InterruptedException {
			
			new Thread(new Runnable() {
				
				@Override
				public void run() {
					final int support = calculateSupportLERS(actionTo,decisionTo);//<----------------------Change if algorithm changes
					
					if(support!=0){
						final String oldConfidence = String.valueOf((Integer.parseInt(calculateConfidenceLERS(actionFrom,decisionFrom)) * 
															Integer.parseInt(calculateConfidenceLERS(actionTo,decisionTo))/100));
						final String newConfidence = calculateConfidenceLERS(actionTo,decisionTo);
						
						if((support >= minSupport && Double.parseDouble(oldConfidence) >= minConfidence ) && !oldConfidence.equals("0") 
								&& !newConfidence.equals("0")){
							
							if(actions!=null){
													
								Collections.sort(actions);
								if(!actionRules.contains(actions)){
									actionRules.add(actions);
													
												try {
													Text key = new Text(rule + " ==> " + "(" + decisionAttribute + "," + decisionFrom + "->" + decisionTo +")");
													Text value = new Text("  "
															+ " [Support:- " + support + " & New Confidence:- " + newConfidence +"% & Old Confidence:- " + oldConfidence + "%]");
													context.write(key , value);
													
												} 
												catch (IOException e) {
													// TODO Auto-generated catch block
													e.printStackTrace();
												} catch (InterruptedException e) {
													// TODO Auto-generated catch block
													e.printStackTrace();
												}
										
								}
							}						
							
						}
						
					}
				}
			}){}.start();
				
			
		}
		
		public boolean isStable(String value){
			if(stableAttributeValues.containsAll(distinctAttributeValues.get(value)))
				return true;
			else return false;
		}
		
	}
	
	public static class JobReducer extends Reducer<Text,Text,Text,Text>{

		@Override
		protected void reduce(Text key, Iterable<Text> values,
				Reducer<Text, Text, Text, Text>.Context context)
				throws IOException, InterruptedException {
			
//			Configuration job = context.getConfiguration();
//			String mapCount = job.get("mapred.map.tasks");
			
			Text finalSum = new Text();
			int sum = 0;
			
			for (Text string : values) {
				sum += 1;
			}
			
			finalSum.set(String.valueOf(sum));
			context.write(key,finalSum);
//			context.write(key, new Text(mapCount));
			
//			super.reduce(key, values, context);
		}
		
	}
		
}