package edu.uncc.kdd;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Scanner;
import java.util.Set;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;


public class Rules {
	static Scanner input;

	static String dirName="C:\\Temp_Grp3";
	public static List<String> attributeNames = new ArrayList<String>();
	public static List<String> stableAttributes = new ArrayList<String>();
	public static List<String> flexibleAttributes = new ArrayList<String>();

	public static Set<String> stableAttr = new HashSet<String>();

	public static String decisionAttribute,decisionFrom,decisionTo;

	public static ArrayList<ArrayList<String>> new_data = new ArrayList<ArrayList<String>>();

	static Map<String, HashSet<String>> distinctAttributeValues = new HashMap<String, HashSet<String>>();
	static Map<HashSet<String>, HashSet<String>> attributeValues = new HashMap<HashSet<String>, HashSet<String>>();
	static Map<HashSet<String>, HashSet<String>> reducedAttributeValues = new HashMap<HashSet<String>, HashSet<String>>();
	static Map<String, HashSet<String>> decisionValues = new HashMap<String, HashSet<String>>();
	static Map<ArrayList<String>, HashSet<String>> markedValues = new HashMap<ArrayList<String>, HashSet<String>>();
	public static Map<ArrayList<String>,String> certainRules = new HashMap<ArrayList<String>,String>();
	public static Map<ArrayList<String>,HashSet<String>> possibleRules = new HashMap<ArrayList<String>,HashSet<String>>();


	static File file ;
	static File actionFile ;
	static File ontologyFile ;
	static BufferedWriter certainRulesOutput;
	static BufferedWriter actionRulesoutput;
	static BufferedWriter ontooutput;


	/**
	 * Create a folder
	 */
	private static void createFolder(){
		new File(dirName).mkdirs();
	}

	//Reading attributes and data
	static void readAttributes(String filePath) {

		createFolder();
		file = new File(dirName+"\\"+"CertainRulesOutput.txt");

		try {
			certainRulesOutput = new BufferedWriter(new FileWriter(file));
		} catch (IOException e1) {
			e1.printStackTrace();
		}

		try {
			File file1 = new File(filePath);
			input = new Scanner(file1);
			while (input.hasNext()) {
				attributeNames.add(input.next().trim());
			}

			for(int i=0; i<attributeNames.size() ;i++){
				if(i<attributeNames.size()-1) Application.stableModel.addElement(attributeNames.get(i));
				else {
					decisionAttribute = attributeNames.get(i);
					Application.decisionModel.addElement(attributeNames.get(i));
					Application.decisionModel.setElementAt(attributeNames.get(i), 1);  //addElement(attributeNames.get(i));
				}
			}
		} catch (FileNotFoundException e) {
			System.out.println("File Not Found");
			e.printStackTrace();
		}

	}

	/**
	 * Read data file
	 * @param filePath
	 */
	static void readData(String filePath) {
		try {
			File file = new File(filePath);
			input = new Scanner(file);
			int lineNo = 0;

			while(input.hasNextLine()){

				String inputStr = input.nextLine();
				if(!inputStr.contains("?")){
					String[] lineData ;
					if(inputStr.contains(",")) lineData= inputStr.split(",");
					else lineData= inputStr.split("\\s+");

					String key;
					if(lineNo==0 && attributeNames.size()<lineData.length) attributeNames.add("Classification");
					lineNo++;
					ArrayList<String> tempList = new ArrayList<String>();
					ArrayList<String> newTempList = new ArrayList<String>();
					HashSet<String> set;

					for (int i=0;i<lineData.length;i++) {
						String currentAttributeValue = lineData[i];
						String attributeName = attributeNames.get(i);
						key = attributeName +"_"+currentAttributeValue;
						newTempList.add(key);
						tempList.add(currentAttributeValue);

						if (i==lineData.length-1) setMap(decisionValues,lineData[i],key,lineNo);
						else{
							HashSet<String> mapKey = new HashSet<String>();
							mapKey.add(key);
							setMap(attributeValues,lineData[i],mapKey,lineNo);
						}

						if (distinctAttributeValues.containsKey(attributeName)) {
							set = distinctAttributeValues.get(attributeName);
							set.add(key);
						}else 
							set = new HashSet<String>();

						set.add(key);
						distinctAttributeValues.put(attributeName, set);
					}

					//data.add(tempList);
					new_data.add(newTempList);
				}
			}
			distinctAttributeValues.size();
			//data.size();
		} catch (FileNotFoundException e) {
			System.out.println("File Not Found");
			e.printStackTrace();
		}
	}


	/**
	 * Find Possible and Certain Rules
	 * @param confidenceThreshold
	 */
	static void findRules(float confidenceThreshold) {
		int loopCount = 0;

		while(!attributeValues.isEmpty()){

			for (Map.Entry<HashSet<String>, HashSet<String>> set : attributeValues.entrySet()) {
				ArrayList<String> setKey = new ArrayList<String>();
				setKey.addAll(set.getKey());

				if (set.getValue().isEmpty()) continue;
				else{
					for(Map.Entry<String, HashSet<String>> decisionSet : decisionValues.entrySet()){
						if(decisionSet.getValue().containsAll(set.getValue())){

							certainRules.put(setKey, decisionSet.getKey());
							markedValues.put(setKey, set.getValue());
							break;
						}
					}
				}

				if(!markedValues.containsKey(setKey)){
					HashSet<String> possibleRulesSet = new HashSet<String>();
					for(Map.Entry<String, HashSet<String>> decisionSet : decisionValues.entrySet()){
						int count =0;
						HashSet<String> values=set.getValue();
						Iterator<String> it = values.iterator();
						while(it.hasNext()){
							String val = it.next();
							if(decisionSet.getValue().contains(val)) count++;
						}
						float confidence = (float)count / (float)values.size();
						if(confidence>= confidenceThreshold) possibleRulesSet.add(decisionSet.getKey());
					}
					if(possibleRulesSet.size()>0) possibleRules.put(setKey, possibleRulesSet);
				}
			}
			removeMarkedValues();
			printCertainRulesMap(certainRules);
			combinePossibleRules();
		}
	}

	static Map<String,ArrayList<String>> atomicActionMap = new HashMap<String, ArrayList<String>>();

	/**
	 * Generate Action Rules as per support and confidence
	 * @param supportThreshold
	 * @param confiThreshold
	 */
	static void generateActionRules(int supportThreshold, float confiThreshold) {
		String rule ="";
		//int loop1Count = 0;

		String action="A";
		int count =0;
		try {
			actionFile = new File(dirName+"\\"+"ActionRulesOutput.txt");
			actionRulesoutput = new BufferedWriter(new FileWriter(actionFile));
			actionRulesoutput.write("\n Action Rules: \n");

			for(Map.Entry<ArrayList<String>, String> certainRules1 : certainRules.entrySet()){
				ArrayList<String> list = new ArrayList<String>();
				if(certainRules1.getValue().equals(decisionFrom)){
					for(Map.Entry<ArrayList<String>, String> certainRules2 : certainRules.entrySet()){
						if((certainRules1.getKey().equals(certainRules2.getKey())) || (!certainRules2.getValue().equals(decisionTo))
								|| (certainRules1.getKey().size()>=certainRules2.getKey().size() )) continue;
						else{
							String primeAttribute = "";
							ArrayList<String> checkCertainValues1 = certainRules1.getKey();
							ArrayList<String> checkCertainValues2 = certainRules2.getKey();
							//int support1 = calculateSupportLERS(checkCertainValues1,certainRules1.getValue());
							//int support2 = calculateSupportLERS(checkCertainValues2,certainRules2.getValue());
							//int minSup = support1<support2?support1:support2;

							//float conf1 = findConfidence(support1, checkCertainValues1);
							//float conf2 = findConfidence(support2, checkCertainValues2);
							//float confidence = calculateConfidence(conf1, conf2);

	//						if(minSup>=supportThreshold || confidence>=confiThreshold){

								for(String value1 : checkCertainValues1){
									if(stableAttributes.contains(value1)) continue;
									else{
										primeAttribute = checkAttribute(value1);
										for(String value2 : checkCertainValues2){
											if(stableAttributes.contains(value2) || !(checkAttribute(value2).equals(primeAttribute)))
												rule=formRule(rule,value2,value2);
											else if(checkAttribute(value2).equals(primeAttribute))
												rule=formRule(rule,value1,value2);
										}
									}
								}
//							}

							if(rule.indexOf(primeAttribute)!=-1 && !primeAttribute.isEmpty()){
								count++;
								if(rule.contains("^")){

									String[] atomicActionArray = rule.split("\\^");
									int len = atomicActionArray.length;
									ArrayList<String> mapValue ;

									for(int i=0; i<len; i++){
										if(atomicActionMap.containsKey(atomicActionArray[i])) {
											mapValue = atomicActionMap.get(atomicActionArray[i]);
											mapValue.add(action+count);
										} 
										else {
											mapValue = new ArrayList<String>();
											mapValue.add(action+count);
										}
										atomicActionMap.put(atomicActionArray[i],mapValue);
									}

								}
								actionRulesoutput.write(rule + " ==> " + "(" + decisionFrom + "->" + decisionTo +")");
								actionRulesoutput.write("\n");
							}
							rule = "";
						}
						//loop2Count++;
					}
				}else continue;
			}
		}catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		finally{
			try {
				try {
					writeIntoJSON();
				} catch (JSONException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				actionRulesoutput.close();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}


	//***************HELPER METHODS****************************




	private static void writeIntoJSON() throws JSONException, IOException{

		ontologyFile = new File(dirName+"\\"+"ontology.txt");
		ontooutput = new BufferedWriter(new FileWriter(ontologyFile));
		
		Map<String,Integer> countMap= new HashMap<String,Integer>();
		JSONArray jtarget = new JSONArray();
		JSONArray ja = new JSONArray();

		Set<String> set = new HashSet<String>();
		int size = atomicActionMap.size();
		int count=0;
		int counter=0;

		Iterator<Entry<String, ArrayList<String>>> iterator1 = atomicActionMap.entrySet().iterator();
		while(iterator1.hasNext()) {
			Entry<String, ArrayList<String>> entry = iterator1.next();
			for(String str:entry.getValue()){
				set.add(str);	
			}
		}

		JSONObject jobj = new JSONObject();
		jobj.put("name", "A");
		jobj.put("size", "60");
		ja.put(jobj);

		for(String str:set){
			count++;
			JSONObject jo = new JSONObject();
			jo.put("name", str);
			jo.put("size", "30");
			ja.put(jo);

			counter++;
			JSONObject jtar = new JSONObject();
			jtar.put("source", 0);
			jtar.put("target", count);
			jtarget.put(jtar);

		}
		Iterator<Entry<String, ArrayList<String>>> iterator2 = atomicActionMap.entrySet().iterator();
		while(iterator2.hasNext()) {
			Entry<String, ArrayList<String>> entry = iterator2.next();

			JSONObject jo = new JSONObject();
			jo.put("name", entry.getKey());
			jo.put("size", "10");
			ja.put(jo);
			countMap.put(entry.getKey(), counter + 1);
			counter++;
		}


		for(String str:set){

			List<String> keys =getValueFromKeys(atomicActionMap, str);
			String subStr=str.substring(1);
			for(String key : keys){
				JSONObject jtar = new JSONObject();
				jtar.put("source", Integer.valueOf(subStr));
				jtar.put("target", countMap.get(key));
				jtarget.put(jtar);
			}
		}

		JSONObject mainObj = new JSONObject();
		mainObj.put("nodes", ja);
		mainObj.put("links", jtarget);

		ontooutput.write(mainObj.toString(2) );
		//System.out.printf( "%s", mainObj.toString(2) );
		ontooutput.close();
		
	}


	private static List<String> getValueFromKeys(Map<String, ArrayList<String>> atomicActionMap, String str ) {

		List<String> values=  new ArrayList<String>();
		Iterator<Entry<String, ArrayList<String>>> itr = atomicActionMap.entrySet().iterator();
		while(itr.hasNext()){
			Entry<String, ArrayList<String>> entry = itr.next();
			if(entry.getValue().contains(str)) values.add(entry.getKey());
		}
		return values;

	}

	private static void removeMarkedValues() {
		for(Map.Entry<ArrayList<String>, HashSet<String>> markedSet : markedValues.entrySet()){
			attributeValues.remove(new HashSet<String>(markedSet.getKey()));
		}
	}

	private static void combinePossibleRules() {
		Set<ArrayList<String>> keySet = possibleRules.keySet();
		ArrayList<ArrayList<String>> keyList = new ArrayList<ArrayList<String>>();
		keyList.addAll(keySet);



		for(int i = 0; i<possibleRules.size(); i++){
			for(int j = (i+1); j<possibleRules.size(); j++){
				HashSet<String> combinedKeys = new HashSet<String>(keyList.get(i));
				combinedKeys.addAll(new HashSet<String>(keyList.get(j)));

				if(!checkSameGroup(combinedKeys)){
					combineAttributeValues(combinedKeys);
				}
			}
		}
		possibleRules.clear();
		removeRedundantValues();
		clearAttributeValues();

	}



	/**
	 * get Attribute
	 * @param value1
	 * @return
	 */
	private static String checkAttribute(String value1) {
		for(Map.Entry<String, HashSet<String>> entryValue : distinctAttributeValues.entrySet()){
			if(entryValue.getValue().contains(value1)) return entryValue.getKey();
		}
		return null;
	}

	/**
	 * Form Rule
	 * @param rule
	 * @param value1
	 * @param value2
	 * @return
	 */
	private static String formRule(String rule,String value1, String value2) {
		if(!rule.isEmpty())
			rule += "^";

		if(value1.equals(value2))
		{
			String[] val = value1.split("_");
			if(stableAttributes.contains(value2)) rule += "(" + val[0] + "=" + val[1] + ")";
			else rule += "(" + val[0] + ",->" + val[1] + ")";
		}
		else
		{
			String[] val1 = value1.split("_");
			String[] val2 = value2.split("_");
			rule += "(" + val1[0] +","+val1[1] +"->" + val2[1] + ")";
		}

		return rule;
	}

	/**
	 * Calculate Support
	 * @param key
	 * @param value
	 * @return
	 */
	private static int calculateSupportLERS(ArrayList<String> key, String value) {
		ArrayList<String> tempList = new ArrayList<String>();
		int count = 0;

		tempList.addAll(key);
		if (!value.isEmpty()) tempList.add(value);

		for (ArrayList<String> data : new_data) {
			if (data.containsAll(tempList)) count++;
		}
		return count;
	}

	/**
	 * Using conf1*conf2
	 * @param conf1
	 * @param conf2
	 * @return
	 */
	private static float calculateConfidence(float conf1, float conf2) {
		float confidence = conf1*conf2;
		return confidence;
	}

	/**
	 * Calculate confidence
	 * @param support
	 * @param key
	 * @return
	 */
	private static float findConfidence(int support, ArrayList<String> key) {

		int den = calculateSupportLERS(key, "");
		if (den == 0) return -1;
		float confidence = support/den;
		return confidence;
	}


	/**
	 * 
	 * @param value
	 */
	private static void printCertainRulesMap(Map<ArrayList<String>, String> value) {
		try {

			certainRulesOutput.write("\nCertain Rules: \n");
			for(Map.Entry<ArrayList<String>,String> set : value.entrySet()){
				certainRulesOutput.write(set.getKey().toString() + " -> " + set.getValue());
				certainRulesOutput.write("\n");
			}
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	private static void setMap(Map<String, HashSet<String>> values,
			String string, String key, int lineNo) {
		HashSet<String> tempSet;

		if (values.containsKey(key)) tempSet = values.get(key);						
		else tempSet = new HashSet<String>();

		tempSet.add(""+lineNo);
		values.put(key, tempSet);
	}

	private static void setMap(Map<HashSet<String>, HashSet<String>> values,
			String string, HashSet<String> key, int lineNo) {
		HashSet<String> tempSet;

		if (values.containsKey(key)) tempSet = values.get(key);					
		else tempSet = new HashSet<String>();

		tempSet.add(""+lineNo);
		values.put(key, tempSet);
	}

	static void setStableAttributes(List<String> attributes, List<String> selectedStableValue) {

		for(String str: selectedStableValue){
			String userStableAttribute = str;
			if(userStableAttribute.length()!=0 && checkValid(attributes,userStableAttribute))
				stableAttributes.addAll(distinctAttributeValues.get(userStableAttribute));
		}
		stableAttr.clear();
		stableAttr.addAll(selectedStableValue);
	}


	private static boolean checkValid(List<String> attributes,String userStableAttribute) {
		if(attributes.contains(userStableAttribute)) return true;
		return false;
	}

	static void setDecisionAttribute(List<String> attributes) {
		if (checkValid(attributes,decisionAttribute)) {
			HashSet<String> decisionValues = distinctAttributeValues.get(decisionAttribute);
			Application.decisionLabel.setText(decisionValues.toString());
		}
	}

	/**
	 * Set Flexible Attributes
	 * Those which are nor stable and decision attributes are flexible
	 */
	static void setFlexibleAttributes(){
		attributeNames.remove(decisionAttribute);
		attributeNames.removeAll(stableAttr);
		flexibleAttributes = attributeNames;
	}


	/**
	 * Set decision values
	 */
	static void getDecisionFromValue() {
		HashSet<String> decisionValues = distinctAttributeValues.get(decisionAttribute);
		decisionFrom = Application.decisionFromField.getText();
		decisionTo = Application.decisionToField.getText();
		if (decisionValues.contains(decisionFrom)) return;
	}

	private static boolean checkSameGroup(HashSet<String> combinedKeys) {
		for(Map.Entry<String, HashSet<String>> singleAttribute : distinctAttributeValues.entrySet()){
			if(singleAttribute.getValue().containsAll(combinedKeys)) return true;
		}
		return false;
	}

	private static void combineAttributeValues(HashSet<String> combinedKeys) {
		HashSet<String> combinedValues = new HashSet<String>();

		for(Map.Entry<HashSet<String>, HashSet<String>> attributeValue : attributeValues.entrySet()){
			if(combinedKeys.containsAll(attributeValue.getKey())){
				if(combinedValues.isEmpty()) combinedValues.addAll(attributeValue.getValue());
				else combinedValues.retainAll(attributeValue.getValue());
			}
		}
		reducedAttributeValues.put(combinedKeys, combinedValues);
	}

	private static void removeRedundantValues() {
		HashSet<String> mark = new HashSet<String>();

		for(Map.Entry<HashSet<String>, HashSet<String>> reducedAttributeValue : reducedAttributeValues.entrySet()){
			for(Map.Entry<HashSet<String>, HashSet<String>> attributeValue : attributeValues.entrySet()){
				if(attributeValue.getValue().containsAll(reducedAttributeValue.getValue()) || reducedAttributeValue.getValue().isEmpty())
					mark.addAll(reducedAttributeValue.getKey());
			}
		}
		reducedAttributeValues.remove(mark);
	}

	private static void clearAttributeValues() {
		attributeValues.clear();
		for(Map.Entry<HashSet<String>, HashSet<String>> reducedAttributeValue : reducedAttributeValues.entrySet()){
			attributeValues.put(reducedAttributeValue.getKey(), reducedAttributeValue.getValue());
		}
		reducedAttributeValues.clear();
	}

}
